/*
 * Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#ifndef __VIRTNET_DPA_REFCOUNT_H__
#define __VIRTNET_DPA_REFCOUNT_H__

#include <stdint.h>
#include <stdbool.h>
#include <libutil/atomic.h>
#include <virtnet_dpa_util.h>

#ifdef __cplusplus
extern "C" {
#endif

struct refcount {
	atomic32_t cnt;
	uint8_t valid;		/* when not set, try_get will abort */
};

static inline void
virtnet_dpa_refcount_init(struct refcount *refcount)
{
	atomic32_set(&refcount->cnt, 1);
	WRITE_ONCE_SCALAR(refcount->valid, 1);
}

/**
 * refcount_try_get - increase a refcount unless valid is not set
 * @refcount: the refcount tried to get
 *
 * Return: false if the passed valid is not set, true otherwise
 */
static inline bool
virtnet_dpa_refcount_try_get(struct refcount *refcount)
{
	/* Publish the value so that, refcount_put() can wait for this acquire
	 * to eventually drop.
	 */
	atomic32_inc(&refcount->cnt);
	/* Ensure that valid flag check occurs after setting the refcount,
	 * otherwise virtnet_dpa_refcount_put_wait() changes to valid set is
	 * not visible.
	 */
	fence_rw();
	if (!READ_ONCE_SCALAR(refcount->valid)) {
		atomic32_dec(&refcount->cnt);
		return false;
	}
	return true;
}

static inline void
virtnet_dpa_refcount_put(struct refcount *refcount)
{
	atomic32_dec(&refcount->cnt);
}

static inline void
virtnet_dpa_refcount_put_wait(struct refcount *refcount)
{
	/* Publish to all acquire routines that any refcount taken
	 * might have to be dropped.
	 */
	WRITE_ONCE_SCALAR(refcount->valid, 0);

	virtnet_dpa_refcount_put(refcount);

	/* Ensure that valid is visible to the try_get() before the
	 * refcount is dropped.
	 */
	fence_rw();
	/* At this point refcount dropped to zero. And no new thread can
	 * acquire the it successfully.
	 */
	while (atomic32_read(&refcount->cnt));
}

#ifdef __cplusplus
}
#endif

#endif
